(* 代数的等価変換 *)

open Global
open ILang

let log2 i =
  let rec iter i j k =
    if j > i then -1
    else if j = i then k
    else iter i (j*2) (k+1)
  in
  iter i 1 0

let rec g e =
  match e with
  | Op(Label "+", [x; Int 0]) | Op(Label "+", [Int 0; x])
  | Op(Label "+.", [x; Float 0.]) | Op(Label "+.", [Float 0.; x])
  | Op(Label "-", [x; Int 0]) | Op(Label "-.", [x; Float 0.])
  | Op(Label "*.", [x; Float 1.]) | Op(Label "*.", [Float 1.; x])
  | Op(Label "/.", [x; Float 1.]) -> V x
  | Op(Label "*", [x; Int 0]) | Op(Label "*", [Int 0; x]) -> V(Int 0)
  | Op(Label "-", [x; y]) when x = y -> V(Int 0)
  | Op(Label "-.", [x; y]) when x = y -> V(Float 0.)
  | Op(Label "*", [x; Int i]) | Op(Label "*", [Int i; x]) ->
      let j = log2 i in
      if j > 0 then Op(Label "lsl", [x; Int j])
      else if j = 0 then V x
      else e
  | Op(Label "/", [x; Int i]) ->
      let j = log2 i in
      if j > 0 then Op(Label "asr", [x; Int j])
      else if j = 0 then V x
      else e
  | Op(Label "/.", [x; Float a]) -> (* 特別に許されるらしい *)
      Op(Label "*.", [x; Float (1. /. a)])
  | If(c,x,y,e1,e2) -> If(c,x,y, g e1, g e2)
  | Let(x,e1,e2) -> Let(x, g e1, g e2)
  | LetRec(x,ys,e1,e2) -> LetRec(x,ys, g e1, g e2)
  | _ -> e

let f e =
  gl_funs := List.map (fun x -> { x with body = g x.body }) !gl_funs;
  g e
