(* 組み込み関数の類 *)

open Id
open Type

let funs_type_alist = let tv = new_tv () in
[["+"; "-"; "*"; "/"; "%"; "lsl"; "asr"], Tfun([Tint; Tint], Tint);
 ["+."; "-."; "*."; "/."], Tfun([Tfloat; Tfloat], Tfloat);
 ["0-"], Tfun([Tint], Tint); (* 単項マイナス *)
 ["0-."; "sqrt"; "floor"], Tfun([Tfloat], Tfloat);
 ["="; "<>"; "=="; "<"; "<="; ">"; ">="], (* != はパーサの都合で略 *)
 ForAll([tv], Tfun([Tvar tv; Tvar tv], Tbool));
 ["xor"; "xnor"; "&"; "or"], Tfun([Tbool; Tbool], Tbool); (* call by value *)
 ["not"], Tfun([Tbool], Tbool);
 ["print_char"], Tfun([Tint], Tunit);
 ["read_char"], Tfun([Tunit], Tint);
 ["create_array"], ForAll([tv], Tfun([Tint; Tvar tv], Tarray(Tvar tv)));
 ["<-"], ForAll([tv], Tfun([Tarray(Tvar tv); Tint; Tvar tv], Tunit));
 ["->"], (* 配列の参照 *)
 ForAll([tv], Tfun([Tarray(Tvar tv); Tint], Tvar tv));
 ["bits_of_float"], Tfun([Tfloat], Tint);
 ["float_of_bits"], Tfun([Tint], Tfloat);
]

(* その他
   "t->", "t<-", ".tuple", ".fill_array", ".call_closure" をこっそり使う *)

let fun_idents =
  List.flatten
    (List.map
       (fun (xs,t) -> List.map (fun x -> new_id_t x t) xs)
       funs_type_alist)

let side_effective_funs = (* 副作用があるかもしれない関数 *)
  ["print_char"; "read_char"; "<-"; ".fill_array"; ".call_closure"]

let non_pure_funs =
  "->"::"create_array"::side_effective_funs
