(* クロージャ変換 *)
(* 1. 直接呼出し、クロージャなし
   2. 直接呼出し、クロージャあり
   3. クロージャ呼出し *)

open Global
open ILang

let result = ref S2.empty (* クロージャが作られる関数の集合 *)

(* env : 関数 -> (クロージャまたはUnit) *)
let rec g env e =
  match e with
  | LetRec(x,ys,e1,e2) ->
      let l = Id.to_string x in
      let funs_bak = !gl_funs in
      (* x の中ではクロージャがいらないと仮定して変換してみる *)
      let env_ = M.add x Unit env in
      let e1_ = g env_ e1 in
      let fv1 = S.remove_list ys (fv e1_) in
      let fv_, env_, e1_ =
        if S.is_empty fv1 then [], env_, e1_ else
        (* 仮定が間違っていたのでやり直す *)
        let env_ = M.add x (Var x) env in
        gl_funs := funs_bak;
        let e1_ = g env_ e1 in
        Format.eprintf "free variable(s) %s found in function %s@."
          (Id.list_to_string (S.elements fv1)) l;
        (S.elements (S.remove x (S.remove_list ys (fv e1_))), env_, e1_)
      in
      let rec iter i fvs =
        match fvs with
        | [] -> e1_
        | z::zs -> Let(z, Op(Label "t->", [Var x; Int i]), iter (i+1) zs)
      in
      gl_funs := {
        name = l;
        args = x::ys;
        body = iter 1 fv_;
      } :: !gl_funs;
      let e2_ = g env_ e2 in
      if fv_ <> [] || S.mem x (fv e2_) then begin
        Format.eprintf "closure for %s created@." l;
        result := S2.add l !result;
        Let(x, Op(Label ".tuple",
                  Label l :: List.map (fun x -> Var x) fv_),
            e2_)
      end else e2_
  | Op(Var x, ys) -> (* 引数リストの先頭をクロージャまたはUnitにする *)
      if M.mem x env then Op(Label (Id.to_string x), M.find x env :: ys)
      else Op(Label "call_closure", Var x :: ys)
  | If(c,x,y,e1,e2) -> If(c,x,y, g env e1, g env e2)
  | Let(x,e1,e2) -> Let(x, g env e1, g env e2)
  | Op(x,ys) -> Op(x,ys)
  | V x -> V x

let f e = g M.empty e
