(* 定数畳み込み *)

open Global
open ILang

let rec g env e =
  match e with
  | If(Eq, Int x, Int y, e1, e2) -> if x = y then g env e1 else g env e2
  | If(LE, Int x, Int y, e1, e2) -> if x <= y then g env e1 else g env e2
  | If(Eq, Float x, Float y, e1, e2) -> if x = y then g env e1 else g env e2
  | If(LE, Float x, Float y, e1, e2) -> if x <= y then g env e1 else g env e2
  | If(_, Unit, _, e1, _) -> g env e1
  | If(c,x,y,e1,e2) ->
      let e1_ = g env e1 and e2_ = g env e2 in
      if SameExp.f e1_ e2_ then e1_
      else If(c,x,y, e1_, e2_)
  | Op(Label "+", [Int i; Int j]) -> V(Int(i + j))
  | Op(Label "-", [Int i; Int j]) -> V(Int(i - j))
  | Op(Label "*", [Int i; Int j]) -> V(Int(i * j))
  | Op(Label "/", [Int i; Int j]) -> V(Int(i / j))
  | Op(Label "lsl", [Int i; Int j]) -> V(Int(i lsl j))
  | Op(Label "asr", [Int i; Int j]) -> V(Int(i asr j))
  | Op(Label "+.", [Float a; Float b]) -> V(Float(a +. b))
  | Op(Label "-.", [Float a; Float b]) -> V(Float(a +. b))
  | Op(Label "*.", [Float a; Float b]) -> V(Float(a +. b))
  | Op(Label "/.", [Float a; Float b]) -> V(Float(a +. b))
  | Op(Label "fabs", [Float a]) -> V(Float(abs_float a))
  | Op(Label "floor", [Float a]) -> V(Float(floor a))
  | Op(Label "sqrt", [Float a]) -> V(Float(sqrt a))
  | Op(Label "float_of_int", [Int i]) -> V(Float(float_of_int i))
  | Op(Label "int_of_float", [Float a]) -> V(Int(int_of_float a))
  | Op(Label "not", [Int i]) -> V(Int(1 - i))
  | Op(Label "xor", [Int i; Int j]) -> V(Int(i lxor j))
  | Op(Label "t->", [Var x; Int i]) when M.mem x env ->
      V(List.nth (M.find x env) i)
  | Op(Label "t->", [Label l; Int i]) ->
      (try
        (match List.nth (Hashtbl.find Global.gl_vars l) i with
        | Var _ -> e
        | o -> V o)
      with Not_found -> e)

  | Let(x, (Op(Label ".tuple", ys) as e1), e2) ->
      Let(x, e1, g (M.add x ys env) e2)
  | Let(x,e1,e2) -> Let(x, g env e1, g env e2)
  | LetRec(x,ys,e1,e2) -> LetRec(x,ys, g env e1, g env e2)
  | _ -> e

let f e =
  gl_funs := List.map
    (fun func -> { func with body = g M.empty func.body })
    !gl_funs;
  g M.empty e
