(* 不要定義削除 *)

open Shared
open Id
open Global
open ILang

let fnames = ref S2.empty

(* 引数の env は副作用がないとわかる関数の集合 *)
let rec g env e =
  match e with
  | V x -> (e, fv0 x, false)
  | Op(Label l, xs) -> fnames := S2.add l !fnames;
      (e, fv1 xs S.empty,
       List.mem l Builtin.side_effective_funs ||
       S2.mem l !SideEffect.result)
  | Op(Var x,ys) -> (e, S.add x (fv1 ys S.empty), not (S.mem x env))
  | Op _ -> assert false
  | If(c,x,y,e1,e2) ->
      let e1_, liv1, se1 = g env e1 in
      let e2_, liv2, se2 = g env e2 in
      (If(c,x,y,e1_,e2_),
       S.union (fv0 x) (S.union (fv0 y) (S.union liv1 liv2)),
       se1 || se2)
  | Let(x,e1,e2) ->
      let e1_, liv1, se1 = g env e1 in
      let e2_, liv2, se2 = g env e2 in
      if se1 || S.mem x liv2 then
        let e_ = match e2_ with
        | V(Var y) when Id.equal x y -> e1_
        | V Unit -> e1_
        | _ -> Let(x,e1_,e2_)
        in
        (e_, S.union liv1 (S.remove x liv2), se1 || se2)
      else begin
        Format.eprintf "%s eliminated@." (Id.to_string x);
        (e2_, liv2, se2)
      end
  | LetRec(x,ys,e1,e2) ->
      let e1_, liv1, se1 = g env e1 in
      let env_ = if se1 then env else S.add x env in
      let e2_, liv2, se2 = g env_ e2 in
      if S.mem x liv2 then
        (LetRec(x,ys,e1_,e2_),
         S.remove x (S.union (S.remove_list ys liv1) liv2),
         se2)
      else begin
        Format.eprintf "%s eliminated@." (Id.to_string x);
        (e2_, liv2, se2)
      end

let f e =
  (* ignore(SideEffect.f e) をここで呼ぶ方が安全ではある *)
  fnames := S2.empty;
  let e,_,_ = g S.empty e in
  let rec iter src dest =
    match List.partition (fun { name = x } -> S2.mem x !fnames) src with
    | [],_ -> dest
    | used,rest -> iter rest
          (List.fold_left
             (fun acc x -> let e,_,_ = g S.empty x.body in
             { x with body = e } :: acc)
             dest used)
  in
  gl_funs := iter !gl_funs [];
  e
