(* 可能なら内部自由変数を外部自由変数にする *)

open Shared
open Id
open ILang
open Global

let rec fv_in_fun e =
  match e with
  | LetRec(f,xs,e1,e2) ->
      S.remove f (S.union (S.remove_list xs (fv e1)) (fv_in_fun e2))
  | If(_,_,_,e1,e2) | Let(_,e1,e2) -> S.union (fv_in_fun e1) (fv_in_fun e2)
  | _ -> S.empty

let is_const = function Var _ -> false | _ -> true

(* env は、配列や組を指すポインタがグローバルになったものの集合 *)
let rec g env e =
  match e with
  | Let(x, Op(Label ".tuple", ys), e2) ->
      let l = Id.to_string x in
      Hashtbl.add gl_tenv l x.typ;
      let rec iter i = function
        | [] -> ([], g env (Alpha.g (M.add x (Label l) M.empty) e2))
        | y::ys -> let lst, e_ = iter (i+1) ys in
          if is_const y then (y :: lst, e_)
          else (y :: lst,
                Let(new_id_t "$" Type.Tunit,
                    Op(Label "t<-", [Label l; Int i; y]), e_))
      in
      let init, e_ = iter 0 ys in
      Hashtbl.add gl_vars l init;
      e_
  | Let(x, Op(Label "create_array", [Int i; y]), e2) ->
      let l = Id.to_string x in
      Hashtbl.add gl_tenv l x.typ;
      let e_ = g env (Alpha.g (M.add x (Label l) M.empty) e2) in
      if is_const y then begin
        Hashtbl.add gl_vars l (list_make i y);
        e_
      end else begin
        Hashtbl.add gl_vars l (list_make i y);
        Let(new_id_t "$" Type.Tunit,
            Op(Label ".fill_array", [Label l; Int i; y]), e_)
      end
  | Let(x, e1, e2) when S.mem x (fv_in_fun e2) ->
      (* ポインタをグローバルにする *)
      let l = Id.to_string x in
      Hashtbl.add gl_tenv l (Type.Ttuple[x.typ]);
      Hashtbl.add gl_vars l [Var x];
      let e2_ = g (S.add x env) e2 in
      Let(x, g env e1,
          Let(new_id_t "$" Type.Tunit,
              Op(Label "<-", [Label l; Int 0; Var x]), e2_))
  | Let(x, e1, e2) -> Let(x, g env e1, g env e2)
  | LetRec(f,xs,e1,e2) ->
      let e1_,env_ = S.fold
        (fun y (e,env) -> let y_ = Id.alpha y in
         (Let(y_, Op(Label "t->", [Label (Id.to_string y); Int 0]), e),
          M.add y (Var y_) env))
        (S.inter env (fv e1))
        (e1,M.empty)
      in
      let e1_ = if M.is_empty env_ then e1_ else Alpha.g env_ e1_ in
      LetRec(f, xs, e1_, g env e2)
  | _ -> e

let f e = g S.empty e
