(* エラー表示用 *)

let linenum = ref 1

open Type
let rec pt t =
  match t with
  | Tunit -> "unit"
  | Tint -> "int"
  | Tfloat -> "float"
  | Tbool -> "bool"
  | Tarray t -> pt t ^ " array"
  | Ttuple ts ->
      let rec iter = function
      | [] -> "" | [x] -> x | hd::tl -> hd ^ "*" ^ iter tl
      in
      "(" ^ iter (List.map pt ts) ^ ")"
  | Tfun(ts, t) -> "(" ^ pt (Ttuple ts) ^ " -> " ^ pt t ^ ")"
  | Tvar{contents = None} -> "'a"
  | Tvar{contents = Some t} -> pt t
  | ForAll(_,t) -> pt t

open Printf
open Id
open FlatSyntax
let px x = printf "%s %s\n" (Id.to_string x) (pt x.typ)
let rec p1 e =
  match e with
  | Let(xs,e1,e2) -> List.iter px xs; p1 e2
  | LetRec(x,ys,e1,e2) -> px x; p1 e1; p1 e2
  | If(_,e1,e2) -> p1 e1; p1 e2
  | _ -> ()

open ILang
let rec p2 e =
  match e with
  | Let(x,e1,e2) -> px x; p2 e2
  | LetRec(x,ys,e1,e2) -> px x; p2 e1; p2 e2
  | If(_,_,_,e1,e2) -> p2 e1; p2 e2
  | _ -> ()
