(* tuple の定数番目を取ってくる演算がある
   tuple の作成を Op に統合
   Bool がない
   Label ってなんでしょうね、てへ *)

type operand =
| Unit
| Int of int
| Float of float
| Label of string
| Var of Id.t

type cmp = Eq | LE

type t =
| V of operand
| Op of operand * operand list
| If of cmp * operand * operand * t * t
| Let of Id.t * t * t
| LetRec of Id.t * Id.t list * t * t

(* 自由変数 *)
let fv0 = function Var x -> S.singleton x | _ -> S.empty
let fv1 vs s = List.fold_left (fun s v -> S.union s (fv0 v)) s vs
let rec fv e =
  match e with
  | V x -> fv0 x
  | Op(x,ys) -> fv1 ys (fv0 x)
  | If(c,x,y,e1,e2) ->
      S.union (fv e1) (S.union (fv e2) (S.union (fv0 x) (fv0 y)))
  | Let(x,e1,e2) -> S.union (fv e1) (S.remove x (fv e2))
  | LetRec(f,xs,e1,e2) ->
      S.remove f (S.union (fv e2) (S.remove_list xs (fv e1)))

(* プログラムのサイズ *)
let rec size e =
  match e with
  | If(_,_,_,e1,e2) | Let(_,e1,e2) | LetRec(_,_,e1,e2) -> size e1 + size e2
  | _ -> 1
