{
open Parser
}

(* 正規表現の略記 *)
let space = [' ' '\t']
let eol   = ['\n' '\r'] | "\r\n"
let digit = ['0'-'9']
let lower = ['a'-'z']
let upper = ['A'-'Z']
let special1 = ['&' '|']
let special2 = ['=' '<' '>']
let special3 = ['+' '-']
let special4 = ['*' '/' '%']
let special = special1 | special2 | special3 | special4 | '.'

rule token = parse
| space+
    { token lexbuf }
| eol
    { incr Err.linenum; token lexbuf }
| "(*"
    { comment lexbuf;
      token lexbuf }
| '('
    { LPAREN }
| ')'
    { RPAREN }
| "true"
    { BOOL(true) }
| "false"
    { BOOL(false) }
| digit+
    { INT(int_of_string (Lexing.lexeme lexbuf)) }
| digit+ ('.' digit*)? (['e' 'E'] ['+' '-']? digit+)?
    { FLOAT(float_of_string (Lexing.lexeme lexbuf)) }
| "if"
    { IF }
| "then"
    { THEN }
| "else"
    { ELSE }
| "let"
    { LET }
| "in"
    { IN }
| "rec"
    { REC }
| ','
    { COMMA }
| '='
    { EQUAL }
| "&&"
    { AMPAMP }
| "||"
    { BARBAR }
| '.'
    { DOT }
| "<-"
    { LESS_MINUS }
| ';'
    { SEMICOLON }
| ";;"
    { SEMISEMI }
| eof
    { EOF }
| special1 special*
    { OP1(Lexing.lexeme lexbuf) }
| special2 special*
    { OP2(Lexing.lexeme lexbuf) }
| special3 special*
    { OP3(Lexing.lexeme lexbuf) }
| special4 special*
    { OP4(Lexing.lexeme lexbuf) }
| (lower|'_') (digit|lower|upper|'_')*
    { IDENT(Lexing.lexeme lexbuf) }
| _
    { failwith
        (Printf.sprintf "unknown token %s near characters %d-%d, line %d"
           (Lexing.lexeme lexbuf)
           (Lexing.lexeme_start lexbuf)
           (Lexing.lexeme_end lexbuf)
           !Err.linenum) }
and comment = parse
| "*)"
    { () }
| "(*"
    { comment lexbuf;
      comment lexbuf }
| eof
    { Format.eprintf "warning: unterminated comment@." }
| eol
    { incr Err.linenum; comment lexbuf }
| _
    { comment lexbuf }
