(* 実アセンブリにかなり近い *)

type t = {
  mutable label : string;
  mutable exp   : elt;
  mutable cont  : t;
}
and elt =
| NOP
| HLT
| MOV of string * string
| GET of string
| PUT of string
| LOAD of string * string * string (* 第3オペランドはラベルとか"0"とか *)
| STORE of string * string * string
| ADDI of string * string * string
| BEI of string * int * string
| BNEI of string * int * string
| BLI of string * int * string
| BGEI of string * int * string
| BE of string * string * string
| BNE of string * string * string
| BL of string * string * string
| BGE of string * string * string
| FBL of string * string * string
| FBGE of string * string * string
| JAL of string
| JUMP of string
| JR of string
| ADD of string * string * string
| SUB of string * string * string
| SLL of string * string * int
| SRR of string * string * int
| AND of string * string * string
| OR of string * string * string
| XOR of string * string * string
| XNOR of string * string * string
| FADD of string * string * string
| FSUB of string * string * string
| FMUL of string * string * string
| FINV of string * string
| FSQRT of string * string

let rec end_node = {
  label = "hlt";
  exp = HLT;
  cont = end_node;
}

type fundef = {
  mutable fname : string;
  mutable entry : t;
  mutable fbody : t;
  mutable exit : t;
  mutable stk : int;
}

let nolabel = ""

let make e c = { label = Shared.new_str (); exp = e; cont = c }

let f2b a = Int32.bits_of_float (a *. 2.0)

open Printf
let str exp =
  let (!) x =
    if TreeAsm.is_reg x then let y = String.copy x in y.[0] <- 'r'; y
    else x
  in
  match exp with
  | NOP -> "nop"
  | HLT -> "hlt"
  | MOV(x,y) -> sprintf "addi %s %s 0" !x !y
  | GET x -> sprintf "get %s" !x
  | PUT x -> sprintf "put %s" !x
  | LOAD(x,y,z) -> sprintf "load %s %s %s" !x !y !z
  | STORE(x,y,z) -> sprintf "store %s %s %s" !x !y !z
  | ADDI(x,y,z) -> sprintf "addi %s %s %s" !x !y !z
  | BEI(x,i,z) -> sprintf "bei %s %d %s" !x i !z
  | BNEI(x,i,z) -> sprintf "bnei %s %d %s" !x i !z
  | BLI(x,i,z) -> sprintf "bli %s %d %s" !x i !z
  | BGEI(x,i,z) -> sprintf "bgei %s %d %s" !x i !z
  | BE(x,y,z) -> sprintf "be %s %s %s" !x !y !z
  | BNE(x,y,z) -> sprintf "bne %s %s %s" !x !y !z
  | BL(x,y,z) -> sprintf "bl %s %s %s" !x !y !z
  | BGE(x,y,z) -> sprintf "bge %s %s %s" !x !y !z
  | FBL(x,y,z) -> sprintf "fbl %s %s %s" !x !y !z
  | FBGE(x,y,z) -> sprintf "fbge %s %s %s" !x !y !z
  | JAL x -> sprintf "jal %s" !x
  | JUMP x -> sprintf "jump %s" !x
  | JR x -> sprintf "jr %s" !x
  | ADD(x,y,z) -> sprintf "add %s %s %s" !x !y !z
  | SUB(x,y,z) -> sprintf "sub %s %s %s" !x !y !z
  | SLL(x,y,i) -> sprintf "sll %s %s %d" !x !y i
  | SRR(x,y,i) -> sprintf "srr %s %s %d" !x !y i
  | AND(x,y,z) -> sprintf "and %s %s %s" !x !y !z
  | OR(x,y,z) -> sprintf "or %s %s %s" !x !y !z
  | XOR(x,y,z) -> sprintf "xor %s %s %s" !x !y !z
  | XNOR(x,y,z) -> sprintf "xnor %s %s %s" !x !y !z
  | FADD(x,y,z) -> sprintf "fadd %s %s %s" !x !y !z
  | FSUB(x,y,z) -> sprintf "fsub %s %s %s" !x !y !z
  | FMUL(x,y,z) -> sprintf "fmul %s %s %s" !x !y !z
  | FINV(x,y) -> sprintf "finv %s %s" !x !y
  | FSQRT(x,y) -> sprintf "fsqrt %s %s" !x !y
