let usage = "Mitou Min-Caml Compiler (C) Eijiro Sumii\n" ^
  "Customized by namasute\n" ^
  "usage: " ^ Sys.argv.(0) ^
  " [-inline m] [-iter n] [-expand-if l] ...filenames without \".ml\"..."

let limit = ref 1000

let compile oc =
  let (>>) f g = fun x -> g (f x) in
  Concat.f >> ToFlatSyntax.f >> Typing.f >> ToILang.f >> Opt1.f !limit >>
  Expansion.f >> Englobal.f >> Closure.f >> Opt2.f !limit >> Opt3.f !limit >>
  ToTreeAsm.f >> RegAlloc.f >> StkAlloc.f >> ToLinearAsm.f >> Peephole.f >>
  Emit.f oc

let compile_files files =
  match files with
  | [] -> prerr_endline usage
  | hd::_ -> let oc = open_out (hd ^ ".s") in
    let defs = List.fold_left
        (fun defs filename ->
          let ic = open_in (filename ^ ".ml") in
          let d = Parser.main Lexer.token (Lexing.from_channel ic) in
          close_in ic;
          d @ defs)
        [] files
    in
    compile oc defs;
    close_out oc

let _ = (* 饳ѥμ¹ԤϤ *)
  let files = ref [] in
  Arg.parse
    [("-inline", Arg.Int(fun i -> Inline.threshold := i),
      "maximum size of functions inlined");
     ("-iter", Arg.Int(fun i -> limit := i),
      "maximum number of optimizations iterated");
     ("-expand-if", Arg.Int(fun i -> TailIf.threshold := i),
      "undocumented")]
    (fun s -> files := s :: !files)
    usage;
  compile_files !files
