(* 便利な夾雑物 *)

(* id生成器 *)
let counter = ref 0
let new_int () =
  incr counter;
  !counter
let counter = ()

let new_str () =
  "hoge." ^ string_of_int (new_int ())

(* xを全部取り除く *)
let list_remove x lst =
  List.filter (fun y -> x <> y) lst
let list_removeq x lst =
  List.filter (fun y -> x != y) lst

(* l1 - l2 *)
let list_diffq l1 l2 =
  List.filter (fun x -> not (List.memq x l2)) l1

(* n 番目を取り除く *)
let rec list_remove_nth i lst =
  match lst with
  | [] -> failwith "list_remove_nth"
  | hd::tl ->
      if i < 0 then failwith "list_remove_nth"
      else if i = 0 then tl
      else hd :: list_remove_nth (i-1) tl

(* 置換 *)
let list_substq x y lst =
  List.map (fun z -> if x == z then y else z) lst

(* [x;x; ... ;x] *)
let list_make len x =
  let rec make len x lst =
    if len <= 0 then lst
    else make (len-1) x (x::lst)
  in
  make len x []

(* i番目からj番目まで *)
let rec list_sublist from_i to_j lst =
  if from_i > 0 then match lst with
  | hd::tl -> list_sublist (from_i - 1) (to_j - 1) tl
  | [] -> failwith "list_sublist"
  else if from_i = 0 && to_j >= 0 then
    let rec sublist_from_0 to_j lst =
      if to_j > 0 then match lst with
      | hd::tl -> hd :: sublist_from_0 (to_j - 1) tl
      | [] -> failwith "list_sublist"
      else []
    in
    sublist_from_0 to_j lst
  else failwith "list_sublist"

(* l1 のうち l2 に入ってないものを l2 に追加 *)
let list_mergeq l1 l2 =
  List.fold_left (fun l x -> if List.memq x l then l else x::l) l2 l1

(* 今何番目かを引数に加えてfold_left *)
let list_foldli f init lst =
  let rec fold f acc lst i =
    match lst with
    | [] -> acc
    | hd::tl -> fold f (f acc hd i) tl (i+1)
  in
  fold f init lst 0

(* x が l の何番目か *)
let list_index x l =
  let rec iter i x l =
    match l with
    | [] -> raise Not_found
    | hd::tl -> if x = l then i else iter (i+1) x tl
  in
  iter 0 x l
let list_indexq x l =
  let rec iter i x l =
    match l with
    | [] -> raise Not_found
    | hd::tl -> if x == l then i else iter (i+1) x tl
  in
  iter 0 x l
