(* スタック割り当て *)

open TreeAsm

let stk_map = Hashtbl.create 100

let rec ra_save_needed tail e =
  let rec aux tail exp =
    match exp with
    | Call _ -> not tail
    | If(_,_,_,e1,e2) | IfI(_,_,_,e1,e2) | FIf(_,_,_,e1,e2) ->
        (ra_save_needed tail e1 || ra_save_needed tail e2)
    | _ -> false
  in
  match e with
  | Ret exp -> aux tail exp
  | Let(_,exp,e) -> aux false exp || ra_save_needed tail e
  | Forget(_,e) -> ra_save_needed tail e

let ra_save_needed e = ra_save_needed true e

let rec g n e =
  match e with
  | Forget(_, e) -> g n e
  | Ret exp ->
      let exp_, n_ = g_ n exp in
      (Ret exp_, n_)
  | Let(x, exp, e_) ->
      let exp_, n_ = g_ n exp in
      let e__, n__ = g n_ e_ in
      (Let(x, exp_, e__), n__)
and g_ n exp =
  match exp with
  | Save(r, x) ->
      if not (Hashtbl.mem stk_map x) then begin
        Hashtbl.add stk_map x n;
        (St(r, reg_sp, I n), n+1)
      end else
        (Nop, n)
  | Restore x ->
      assert (Hashtbl.mem stk_map x);
      (Op1i(Ld, reg_sp, I (Hashtbl.find stk_map x)), n)
  | If(c,x,y,e1,e2) ->
      let e1_, n1 = g n e1 in
      let e2_, n2 = g n e2 in
      (If(c,x,y,e1_,e2_), n)
  | IfI(c,x,i,e1,e2) ->
      let e1_, n1 = g n e1 in
      let e2_, n2 = g n e2 in
      (IfI(c,x,i,e1_,e2_), n)
  | FIf(c,x,y,e1,e2) ->
      let e1_, n1 = g n e1 in
      let e2_, n2 = g n e2 in
      (FIf(c,x,y,e1_,e2_), n)
  | _ -> (exp, n)

let g e =
  Hashtbl.clear stk_map;
  let e_, _ = g 1 e in
  let s = if ra_save_needed e then
    1 + Hashtbl.fold (fun _ n m -> max n m) stk_map 0
  else
    Hashtbl.fold (fun _ n m -> max (n+1) m) stk_map 0
  in
  (e_, s)

let f xs =
  Format.eprintf "stack allocation...@.";
  List.map (fun { name = x; args = rs; body = e } ->
  let e_, s = g e in
  Hashtbl.clear stk_map;
  { name = x; args = rs; body = e_; stksiz = s }) xs
