(* FlatSyntax.t -> ILang.t *)

open Type
open Id
open FlatSyntax
open ILang

let h v =
  match v with
  | FlatSyntax.Unit -> Unit
  | FlatSyntax.Int i -> Int i
  | FlatSyntax.Float f -> Float f
  | Bool b -> Int(if b then 1 else 0)
  | FlatSyntax.Var x -> Var x

let rec g e =
  match e with
  | FlatSyntax.V v -> V(h v)
  | FlatSyntax.Tuple vs -> Op(Label ".tuple", List.map h vs)
  | FlatSyntax.Op(x, ys) -> Op(h x, List.map h ys)
  | FlatSyntax.If(x, e1, e2) -> If(Eq, h x, Int 0, g e2, g e1)
  | FlatSyntax.Let(_::_::_ as xs, e1, e2) ->
      let t = Type.Ttuple(List.map (fun x -> x.typ) xs) in
      let tuple = new_id_t "tuple" t in
      let rec iter i xs =
        match xs with
        | [] -> g e2
        | x::xs -> Let(x, Op(Label "t->", [Var tuple; Int i]), iter (i+1) xs)
      in
      Let(tuple, g e1, iter 0 xs)
  | FlatSyntax.Let([x], e1, e2) -> Let(x, g e1, g e2)
  | FlatSyntax.Let([], e1, e2) -> Let(new_id_t "unit" Tunit, g e1, g e2)
  | FlatSyntax.LetRec(x, ys, e1, e2) -> LetRec(x, ys, g e1, g e2)

open List

let add_builtin_fundefs e =
  let add1 e x =
    let ts = match x.typ with
      | Tfun(ts,_) | ForAll(_,Tfun(ts,_)) -> ts
      | _ -> assert false
    in
    let ys = map (fun t -> new_id_t "x" t) ts in
    let vs = map (fun y -> Var y) ys in
    let body = match x.name with
      | "=" | "==" -> If(Eq, hd vs, hd (tl vs), V(Int 1), V(Int 0))
      | "<>" -> If(Eq, hd vs, hd (tl vs), V(Int 0), V(Int 1))
      | "<=" -> If(LE, hd vs, hd (tl vs), V(Int 1), V(Int 0))
      | "<" -> If(LE, hd (tl vs), hd vs, V(Int 0), V(Int 1))
      | ">=" -> If(LE, hd (tl vs), hd vs, V(Int 1), V(Int 0))
      | ">" -> If(LE, hd vs, hd (tl vs), V(Int 0), V(Int 1))
      | "0-" -> Op(Label "-", [Int 0; hd vs])
      | "0-." -> Op(Label "-.", [Float 0.; hd vs])
      | l -> Op(Label l, vs)
    in
    LetRec(x, ys, body, e)
  in
  fold_left add1 e Builtin.fun_idents

let f e = add_builtin_fundefs (g e)
