(* 仮想アセンブリ *)

type t =
| Ret of exp
| Let of string * exp * t
| Forget of string * t
and exp =
| Nop
| If of cmp * string * string * t * t
| IfI of cmp * string * int * t * t
| FIf of cmp * string * string * t * t
| St of string * string * imm
| Op2 of op2 * string * string
| Op1i of op1i * string * imm
| Op1 of op1 * string
| Imm of imm
| Rd
| Call of string * string list
| Mov of string
| Save of string * string (* レジスタ、変数 *)
| Restore of string
and op2 =
| Add
| Sub
| And
| Or
| Xor
| Xnor
| FAdd
| FSub
| FMul
| Array (* 長さ、初期値 *)
and op1i =
| AddI
| Sll
| Srr
| Ld
and op1 =
| FInv
| FSqrt
| Wr
and imm =
| I of int
| F of float
| L of string
and cmp =
| Eq | Lt

type func = {
  name : string;
  args : string list;
  body : t;
  stksiz : int;
}

let regs = Array.init 59
    (fun i -> "%" ^ string_of_int (if i < 28 then i+3 else i+4))
let regs_lst = Array.to_list regs
let regs_set = S2.of_list regs_lst
let reg_zero = "%0"
let reg_sp = "%1"
let reg_hp = "%2"
let reg_tmp = "%63"
let reg_ra = "%31"
let reg_dummy = "%_"

let is_reg r = r.[0] = '%'

let rec concat e1 x e2 =
  match e1 with
  | Ret exp -> Let(x, exp, e2)
  | Let(y, exp, e1) -> Let(y, exp, concat e1 x e2)
  | Forget(y, e1) -> Forget(y, concat e1 x e2)

let rec fv e =
  match e with
  | Ret exp -> fv1 exp
  | Let(x, exp, e) -> S2.union (fv1 exp) (S2.remove x (fv e))
  | Forget(_, e) -> fv e
and fv1 exp =
  match exp with
  | Nop | Imm _ | Rd | Restore _ -> S2.empty
  | Op1(_,x) | Op1i(_,x,_) | Mov x | Save(x,_) -> S2.singleton x
  | Op2(_,x,y) | St(x,y,_) -> S2.add x (S2.singleton y)
  | Call(_,xs) -> S2.of_list xs
  | If(_,x,y,e1,e2) | FIf(_,x,y,e1,e2) ->
      S2.add x (S2.add y (S2.union (fv e1) (fv e2)))
  | IfI(_,x,_,e1,e2) -> S2.add x (S2.union (fv e1) (fv e2))
